package labeling;

import java.awt.Canvas;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.geom.Rectangle2D;
import java.io.File;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.TreeMap;

import rectangleSweep.RectangleSweep;
import rectangleSweep.RectangleWithInfo;

//add gurobi library to build path (for ILP solution) 
import gurobi.GRB;
import gurobi.GRBEnv;
import gurobi.GRBException;
import gurobi.GRBLinExpr;
import gurobi.GRBModel;
import gurobi.GRBVar;

//add jts library to build path (for spatial queries) 
import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.Polygon;
import com.vividsolutions.jts.geom.impl.CoordinateArraySequence;
import com.vividsolutions.jts.index.quadtree.Quadtree;
import com.vividsolutions.jts.index.strtree.STRtree;

//add jump library to build path (for import and export of shapefiles)
import com.vividsolutions.jump.feature.AttributeType;
import com.vividsolutions.jump.feature.BasicFeature;
import com.vividsolutions.jump.feature.Feature;
import com.vividsolutions.jump.feature.FeatureCollection;
import com.vividsolutions.jump.feature.FeatureDataset;
import com.vividsolutions.jump.feature.FeatureSchema;
import com.vividsolutions.jump.io.DriverProperties;
import com.vividsolutions.jump.io.ShapefileReader;
import com.vividsolutions.jump.io.ShapefileWriter;

/**
 * @author haunert
 * 
 * When using or adapting this code for your research, please cite:
 * 
 * Jan-Henrik Haunert and Alexander Wolff
 * Beyond Maximum Independent Set: An Extended Integer Programming Formulation for Point Labeling
 * International Journal of Geo-Information, 2017
 */

public class LabelingAlgorithms {
	
    /**
     * scale applied to point set
     */
	public static double scale = 1/2000.0;

	/**
	 * buffer for labels
	 */
	public static double epsilon = 0.5;
	
	/**
	 * distance for the definition of interferences
	 */
	public static double lambda = 4.0;
	
	/**
	 * width of square-shaped region gamma for setting up the density constraint
	 */
	public static double gamma_width = 25.0;
	
	/**
	 * maximum number of labels allowed to intersect any translated copy of gamma
	 */
	public static int k_gamma = 2;
	
	/**
	 * parameter defining the penalty for interferences 
	 */
	public static double alpha = 0.4;
	
	/**
	 * parameter defining whether the 8-pos model is used when setting up an instance from file; if false, the 4-pos model is used
	 */
    public static boolean use_8pos_model = false;
    
	/**
	 * printwriter for logging
	 */
    public static PrintWriter writer;
    	
    /**
     * total time spent for set up of models since last call of startRecording()
     */
    private static long sumTimeSetUp;
    
    /**
     * total time spent for solution of models since last call of startRecording()
     */
    private static long sumTimeILP; 
    
    /**
     * total time spent by rounding heuristic since last call of startRecording()
     */
    private static long sumTimeRounding;
    
    /**
     * sum of objective of all ILP solutions computed since last call of startRecording()
     */
    private static double sumObjILP;
    
    /**
     * sum of objective of all heuristic solutions computed since last call of startRecording()
     */
    private static double sumObjRounding;       
	
    /**
     * total number of instances solved since last call of startRecording()
     */
    private static int numInstances;
    
    public static void startLogging() {
        numInstances = 0;
        sumObjILP = 0.0;
        sumObjRounding = 0.0;
        sumTimeSetUp = 0;
        sumTimeRounding = 0;
        sumTimeILP = 0;
    }
    
    public static void writeSummaryStatistics() {
        writer.println("-------------------------------");
        writer.println("Summary Statistics");
        writer.println("-------------------------------");
        writer.println("number of instances                 : " + numInstances);
        writer.println("average objective with ILP          : " + (sumObjILP / numInstances));
        writer.println("average objective with heuristic    : " + (sumObjRounding / numInstances));
        writer.println("average time for model set up       : " + (sumTimeSetUp / numInstances));
        writer.println("average time for ILP solution       : " + (sumTimeILP / numInstances));
        writer.println("average time for heuristic solution : " + (sumTimeRounding / numInstances));
        writer.println("-------------------------------");
    }
    
	/**
	 * generates and solves an instance from a shapefile with point features 
	 * @param instanceID: the id of the instance
	 * @param useClique: if true, the strong ILP formulation (ILP 2) is used 
	 * @param filename: the shapefile containing the points
	 * @param fontSize: the size of the font 
	 * @throws GRBException
	 */
	public static void solveGeoInstance(int instanceID, boolean useClique, String filename, int fontSize) throws GRBException {
	    
		LinkedList<Point> points = new LinkedList<Point>();
		LinkedList<Label> labelList = new LinkedList<Label>();
		if(filename.endsWith(".shp")) {
            System.out.println("Point set read from file: " + filename);
	        ShapefileReader shp_input = new ShapefileReader();
	        DriverProperties dp = new DriverProperties(filename);
			FeatureCollection myFeatureCollection = null;
			Canvas c = new Canvas();	
			Font tr = new Font("TimesRoman", Font.PLAIN, fontSize);
			FontMetrics fm = c.getFontMetrics(tr);
			
			try {
	            myFeatureCollection = shp_input.read(dp);
	            int counter = 0;
	       
			    String nameColumn = "name";
			    String rankColumn = "rank_max";
			    
			    GeometryFactory gf = new GeometryFactory();
			    
			    for (Iterator<?> i = myFeatureCollection.iterator(); i.hasNext();) {
	                Feature myFeature = (Feature) i.next();
	             
	                String name = (String) myFeature.getAttribute(nameColumn);
	                name = name.trim();
	                int rank = (int) Math.pow((Integer) myFeature.getAttribute(rankColumn) + 1, 2.0);
			        System.out.println(rank);
	                Geometry myGeometry = myFeature.getGeometry();
			        Coordinate[] xyz = myGeometry.getCoordinates();
			        Point p = new Point(xyz[0].x , xyz[0].y );
			        	        
			        Rectangle2D r = fm.getStringBounds(name, c.getGraphics());
			        System.out.println("n "+ name + " n");
			       	System.out.println("r " + r.getWidth() + " " + r.getHeight());
			        
			        Polygon pNE = (Polygon) gf.toGeometry(new Envelope(
							p.getX() - epsilon / scale,
							p.getX() + (r.getWidth() + epsilon) / scale, 
							p.getY() - epsilon / scale,
							p.getY() + (r.getHeight() + epsilon) / scale));       
			        Label ne = new Label(pNE, p, rank, counter++);
			        ne.setText(name);
			        p.addLabel(ne);
			        labelList.add(ne);
			        
			        Polygon pSE = (Polygon) gf.toGeometry(new Envelope(
							p.getX() - epsilon / scale,
							p.getX() + (r.getWidth() + epsilon) / scale,
							p.getY() - (r.getHeight() + epsilon) / scale,
							p.getY() + epsilon / scale));     
			        Label se = new Label(pSE, p, rank, counter++);
			        se.setText(name);
			        p.addLabel(se);
			        labelList.add(se);
			        
			        Polygon pSW = (Polygon) gf.toGeometry(new Envelope(
							p.getX() - (r.getWidth() + epsilon) / scale,
							p.getX() + epsilon / scale,
							p.getY() - (r.getHeight() + epsilon) / scale,
							p.getY() + epsilon  / scale));
			        Label sw = new Label(pSW, p, rank, counter++);
			        sw.setText(name);
			        p.addLabel(sw);
			        labelList.add(sw);
			        
			        Polygon pNW = (Polygon) gf.toGeometry(new Envelope(
							p.getX() - (r.getWidth() + epsilon) / scale,
							p.getX() + epsilon / scale,
							p.getY() - epsilon / scale,
							p.getY() + (r.getHeight() + epsilon) / scale));     
			        Label nw = new Label(pNW, p, rank, counter++);
			        nw.setText(name);
			        p.addLabel(nw);
			        labelList.add(nw);
			        
			        if (use_8pos_model) {
			            Polygon pE = (Polygon) gf.toGeometry(new Envelope(
	                            p.getX() - epsilon / scale,
	                            p.getX() + (r.getWidth() + epsilon) / scale, 
	                            p.getY() - 0.5 * (r.getHeight() + epsilon) / scale,
	                            p.getY() + 0.5 * (r.getHeight() + epsilon) / scale));       
	                    Label le = new Label(pE, p, rank, counter++);
	                    le.setText(name);
	                    p.addLabel(le);
	                    labelList.add(le);
	                    
	                    Polygon pS = (Polygon) gf.toGeometry(new Envelope(
	                            p.getX() - 0.5 * (r.getWidth() + epsilon) / scale,
	                            p.getX() + 0.5 * (r.getWidth() + epsilon) / scale,
	                            p.getY() - (r.getHeight() + epsilon) / scale,
	                            p.getY() + epsilon / scale));     
	                    Label ls = new Label(pS, p, rank, counter++);
	                    ls.setText(name);
	                    p.addLabel(ls);
	                    labelList.add(ls);
	                    
	                    Polygon pW = (Polygon) gf.toGeometry(new Envelope(
	                            p.getX() - (r.getWidth() + epsilon) / scale,
	                            p.getX() + epsilon / scale,
	                            p.getY() - 0.5 * (r.getHeight() + epsilon) / scale,
                                p.getY() + 0.5 * (r.getHeight() + epsilon) / scale));
	                    Label lw = new Label(pW, p, rank, counter++);
	                    lw.setText(name);
	                    p.addLabel(lw);
	                    labelList.add(lw);
	                    
	                    Polygon pN = (Polygon) gf.toGeometry(new Envelope(
	                            p.getX() - 0.5 * (r.getWidth() + epsilon) / scale,
                                p.getX() + 0.5 * (r.getWidth() + epsilon) / scale,
	                            p.getY() - epsilon / scale,
	                            p.getY() + (r.getHeight() + epsilon) / scale));     
	                    Label ln = new Label(pN, p, rank, counter++);
	                    ln.setText(name);
	                    p.addLabel(ln);
	                    labelList.add(ln);
			        }
			        points.add(p);
			    }
			} catch (Exception e) {
				e.printStackTrace();
			}
			
			Label[] labels = new Label[labelList.size()];
			int i = 0;
			for (Label l : labelList) {
				labels[i++] = l;
			}
			
			System.out.println("number of labels: " + labelList.size());
					
			solveInstance(labels, instanceID, useClique);
		}
	}

	/**
	 * Generates and solves a random instance with the 4-position model
	 * @param instanceID: the id of the instance
     * @param useClique: if true, the strong ILP formulation (ILP 2) is used 
	 * @param instance_w: width of the rectangle containing all points
	 * @param instance_h: height of the rectangle containing all points
	 * @param n_points: number of points
	 * @param rect_h: height of label
	 * @param rect_w: width of label
	 * @throws GRBException
	 */
    public static void solveRandomInstance(int instanceID, boolean useClique, double instance_w, double instance_h, int n_points, double rect_h, double rect_w) throws GRBException {
        
        Point[] points = getRandomPointSet(instance_w, instance_h, n_points, instanceID);
        double[] weights = getRandomWeights(n_points, 4, 0.0, 1.0, instanceID + 100);
        Label[] labels = new Label[n_points * 4];
        GeometryFactory gf = new GeometryFactory();
        for (int i = 0; i < n_points; i++) {
            Point p = points[i];
            
            //NE
            Polygon polygonNE = (Polygon) gf.toGeometry(new Envelope(
                    p.getX() - epsilon, 
                    p.getX() + rect_w + epsilon, 
                    p.getY() - epsilon, 
                    p.getY() + rect_h + epsilon)); 
            Label labelNE = new Label(polygonNE, p, weights[i * 4], i * 4);
            labels[i * 4] = labelNE;
            p.addLabel(labelNE);
            
            //NW
            Polygon polygonNW = (Polygon) gf.toGeometry(new Envelope(
                    p.getX() - rect_w - epsilon, 
                    p.getX() + epsilon, 
                    p.getY() - epsilon, 
                    p.getY() + rect_h + epsilon));
            Label labelNW = new Label(polygonNW, p, weights[i * 4 + 1], i * 4 + 1);
            labels[i * 4 + 1] = labelNW;
            p.addLabel(labelNW);
            
            //SW
            Polygon polygonSW = (Polygon) gf.toGeometry(new Envelope(
                    p.getX() - epsilon, 
                    p.getX() + rect_w + epsilon, 
                    p.getY() - rect_h - epsilon, 
                    p.getY() + epsilon)); 
            Label labelSW = new Label(polygonSW, p, weights[i * 4 + 2], i * 4 + 2);
            labels[i * 4 + 2] = labelSW;
            p.addLabel(labelSW);
            
            //SE
            Polygon polygonSE = (Polygon) gf.toGeometry(new Envelope(
                    p.getX() - rect_w - epsilon, 
                    p.getX() + epsilon, 
                    p.getY() - rect_h - epsilon,
                    p.getY() + epsilon));
            Label labelSE = new Label(polygonSE, p, weights[i * 4 + 3], i * 4 + 3);
            labels[i * 4 + 3] = labelSE;
            p.addLabel(labelSE);
        }
                
        solveInstance(labels, instanceID, useClique);
    }
    
    
	
	private static void solveInstance(Label[] labels, int instanceID, boolean useClique) throws GRBException {

        exportLabels("candidates_" + instanceID + ".shp", labels, new HashSet<Label>());    
	    
        long startTimeSetUp = System.currentTimeMillis();
        writer.println("----------------------------");
        writer.println("instance " + instanceID);
        writer.println("useClique " + useClique);
        int n = labels.length;
        
        HashSet<Point> points = new HashSet<Point>();
        for (int i = 0; i < n; i++) {  
            Point p = labels[i].getPoint();
            points.add(p);
        }
                
        // set up STRtree for efficient detection of interferences
        STRtree t = new STRtree();
        for (int i = 0; i < n; i++) {
            t.insert(labels[i].getPolygon().getEnvelopeInternal(), i);
        }
        
        // find interferences
        TreeMap<Interference, Double> interferences = new TreeMap<Interference, Double>();
        for (Point p : points) {
            Envelope env = new Envelope(p.getX() - lambda / scale, p.getX() + lambda / scale, p.getY() - lambda / scale, p.getY() + lambda / scale);
            List interferingLabels = t.query(env);
            for (Object o : interferingLabels) {
                Label receiver = labels[(Integer) o];
                if (receiver.getPoint() != p && p.getDistance(receiver) < lambda / scale && !receiver.getPolygon().contains(p.toJtsPoint())) {
                    for (Label trigger : p.getLabels()) {
                        if (!receiver.getPolygon().intersects(trigger.getPolygon())) {
                            Interference c = new Interference(trigger, receiver);
                            boolean exists = interferences.containsKey(c);
                            if (exists) {
                                interferences.put(c, interferences.get(c) + receiver.getWeight());
                            } else {
                                interferences.put(c, receiver.getWeight());
                            }
                        }
                    }
                }
            }
        }
                
        /*
         * Set up of ILP for exact solution
         */   
        GRBEnv env = new GRBEnv();
        GRBModel model = new GRBModel(env); 
           
        /*
         * Set up of LP for rounding heuristic
         */
        GRBEnv envLP = new GRBEnv();
        GRBModel modelLP = new GRBModel(envLP);
        
        /*
         * definition of x-variables
         */
        GRBVar[] x = new GRBVar[n];
        GRBVar[] xCont = new GRBVar[n];
        for (int i = 0; i < n; i++) {
            x[i] = model.addVar(0, 1, labels[i].getWeight(), GRB.BINARY, "x_" + i);
            xCont[i] = modelLP.addVar(0.0, 1.0, labels[i].getWeight(), GRB.CONTINUOUS, "");     
        }
        
        /*
         * definition of y-variables for interferences
         */
        Interference[] interferencesArray = null;
        GRBVar[] y = null;
        GRBVar[] yCont = null;
        if ( interferences != null ) {
            Object[] o = interferences.keySet().toArray();
            interferencesArray = new Interference[o.length];
            for (int i = 0; i < o.length; i++) interferencesArray[i] = (Interference) o[i];
            y = new GRBVar[interferencesArray.length];
            yCont = new GRBVar[interferencesArray.length];
            for (int i = 0; i < y.length; i++) {
                y[i] = model.addVar(0.0, 1.0, -alpha * interferences.get(interferencesArray[i]), GRB.CONTINUOUS, "y_" + i); //CONTINOUS
                yCont[i] = modelLP.addVar(0.0, 1.0, -alpha * interferences.get(interferencesArray[i]), GRB.CONTINUOUS, "");
            }
        }
                
        model.update();
        modelLP.update();
        model.set(GRB.IntAttr.ModelSense, GRB.MAXIMIZE);
        modelLP.set(GRB.IntAttr.ModelSense, GRB.MAXIMIZE);
        model.update();
        modelLP.update();
              
        System.out.println("number of interferences: " + y.length);
        
        /*
         * polygons defining the arrangements
         */
        LinkedList<RectangleWithInfo<Label>> rectanglesForLabels = new LinkedList<RectangleWithInfo<Label>>();
        LinkedList<RectangleWithInfo<Label>> viewBoxForLabels = new LinkedList<RectangleWithInfo<Label>>();
        Polygon[] viewboxes = new Polygon[labels.length];
        int counter = 0;
        for (Label l : labels) {
            Envelope lEnv = l.getPolygon().getEnvelopeInternal();
            
            RectangleWithInfo<Label> rect = new RectangleWithInfo<Label>(lEnv.getMinX(), lEnv.getMinY(), lEnv.getWidth(), lEnv.getHeight(), "" + l.getIndex(), l);
            rectanglesForLabels.add(rect);
            
            RectangleWithInfo<Label> viewbox = new RectangleWithInfo<Label>(lEnv.getMinX() - 0.5 * gamma_width / scale , lEnv.getMinY() - 0.5 * gamma_width / scale, lEnv.getWidth() + gamma_width / scale, lEnv.getHeight() + gamma_width / scale, "" + l.getIndex(), l);
            viewBoxForLabels.add(viewbox);
            
            viewboxes[counter] = viewbox.getAsPolygon();
            
            counter++;
        }
        exportPolygons("viewboxes.shp", viewboxes, null);
        
        
        /*
         * overlap constraints based on pairs of labels (ILP 1)
         */
        if (!useClique) {
            for (int i = 0; i < n; i++) {   
                List l = t.query(labels[i].getPolygon().getEnvelopeInternal());
                for (Object o : l) {
                    Integer j = (Integer) o;            
                    if (i < j && labels[i].getPolygon().overlaps(labels[j].getPolygon())) {
                        GRBLinExpr myExprEdge = new GRBLinExpr();
                        myExprEdge.addTerm(1.0, x[i]);
                        myExprEdge.addTerm(1.0, x[j]);  
                        
                        GRBLinExpr myExprEdgeCont = new GRBLinExpr();
                        myExprEdgeCont.addTerm(1.0, xCont[i]);
                        myExprEdgeCont.addTerm(1.0, xCont[j]);
                        
                        model.addConstr(myExprEdge, GRB.LESS_EQUAL, 1.0, "");
                        modelLP.addConstr(myExprEdgeCont, GRB.LESS_EQUAL, 1.0, "");
                    }
                }
            }       
        }
        
        /*
         * overlap constraints based on cliques of labels (ILP 2)
         */
        if (useClique) {
            RectangleSweep<Label> rs = new RectangleSweep<Label>(rectanglesForLabels);
            LinkedList<LinkedList<RectangleWithInfo<Label>>> cliques = rs.computeMaximalCliques();
            
            for (LinkedList<RectangleWithInfo<Label>> clique : cliques) {
                GRBLinExpr myExprClique = new GRBLinExpr();
                GRBLinExpr myExprCliqueCont = new GRBLinExpr();
                for (RectangleWithInfo<Label> rwi : clique) {
                    myExprClique.addTerm(1.0, x[rwi.getInfo().getIndex()]);
                    myExprCliqueCont.addTerm(1.0, xCont[rwi.getInfo().getIndex()]);
                }
                
                model.addConstr(myExprClique, GRB.LESS_EQUAL, 1.0 , "");
                modelLP.addConstr(myExprCliqueCont, GRB.LESS_EQUAL, 1.0 , "");
            }
            model.update();
            modelLP.update();
        }
        System.out.println("overlap constraints done");
        
        /*
         * density constraints
         */
        RectangleSweep<Label> rs_vb = new RectangleSweep<Label>(viewBoxForLabels);
        LinkedList<LinkedList<RectangleWithInfo<Label>>> viewboxCliques = rs_vb.computeMaximalCliques();
        
        for (LinkedList<RectangleWithInfo<Label>> clique : viewboxCliques) {
            GRBLinExpr myExprClique = new GRBLinExpr();
            GRBLinExpr myExprCliqueCont = new GRBLinExpr();
            for (RectangleWithInfo<Label> rwi : clique) {
                myExprClique.addTerm(1.0, x[rwi.getInfo().getIndex()]);
                myExprCliqueCont.addTerm(1.0, xCont[rwi.getInfo().getIndex()]);
            }
            
            model.addConstr(myExprClique, GRB.LESS_EQUAL, k_gamma , "");
            modelLP.addConstr(myExprCliqueCont, GRB.LESS_EQUAL, k_gamma , "");
        }
        model.update();
        modelLP.update();
        System.out.println("density constraints done");
        
        /*
         * constraints linking x- and y-variables 
         */
        if (interferencesArray != null) {
            for (int i = 0; i < interferencesArray.length; i++) {
                Interference conf = interferencesArray[i];
                int l1id = conf.getLabel1().getIndex();
                int l2id = conf.getLabel2().getIndex();
               
                GRBLinExpr myExprInterference = new GRBLinExpr();
                myExprInterference.addTerm(1.0, x[l1id]);
                myExprInterference.addTerm(1.0, x[l2id]);
                myExprInterference.addConstant(-1.0);
                model.addConstr(y[i], GRB.GREATER_EQUAL, myExprInterference, "");

                if (alpha == 0.0) { 
                    model.addConstr(y[i], GRB.LESS_EQUAL, x[l1id], "");
                    model.addConstr(y[i], GRB.LESS_EQUAL, x[l2id], "");
                }
                               
                GRBLinExpr myExprInterferenceCont = new GRBLinExpr();
                myExprInterferenceCont.addTerm(1.0, xCont[l1id]);
                myExprInterferenceCont.addTerm(1.0, xCont[l2id]);
                myExprInterferenceCont.addConstant(-1.0);
                modelLP.addConstr(yCont[i], GRB.GREATER_EQUAL, myExprInterferenceCont, "");
                
                if (alpha == 0.0) {
                    modelLP.addConstr(yCont[i], GRB.LESS_EQUAL, xCont[l1id], "");
                    modelLP.addConstr(yCont[i], GRB.LESS_EQUAL, xCont[l2id], "");                  
                }
            }
        }
        
        model.update();
        modelLP.update();
        System.out.println("model done");
        long endTimeSetUp = System.currentTimeMillis();
        sumTimeSetUp += (endTimeSetUp - startTimeSetUp);
        writer.println("Model set up time: " + (endTimeSetUp - startTimeSetUp));
        
        /*
         * compute heuristic solution
         */
        long roundingStartTime = System.currentTimeMillis();
        modelLP.optimize();
        
        System.out.println("Objective of LP: " + modelLP.get(GRB.DoubleAttr.ObjVal));
        writer.println("Objective of LP: " + modelLP.get(GRB.DoubleAttr.ObjVal));
       
        HashSet<Label> roundingSol = roundingHeuristic(labels, viewboxCliques, xCont); 
        
        long roundingEndTime = System.currentTimeMillis();
        sumTimeRounding += (roundingEndTime - roundingStartTime);
        writer.println("Heuristic solution time: " + (roundingEndTime - roundingStartTime));
        
        double w = getObjective(roundingSol, interferences, alpha);
        sumObjRounding += w;
        
        writer.println("Objective of rounding heuristic: " + w);
    
        exportLabels("polygons_h_" + useClique + "_" + instanceID + ".shp", labels, roundingSol);
        exportPoints("points_h_" + useClique + "_" + instanceID + ".shp", points, roundingSol);
        
        /*
         * compute exact solution 
         */
        long ilpStartTime = System.currentTimeMillis();
        model.optimize();
        long ilpEndTime = System.currentTimeMillis();
                
        sumTimeILP += (ilpEndTime - ilpStartTime);   
        sumObjILP += model.get(GRB.DoubleAttr.ObjVal);
        
        writer.println("ILP solution time: " + (ilpEndTime - ilpStartTime));
        writer.println("Objective: " + model.get(GRB.DoubleAttr.ObjVal));
        writer.println("n constraints: " + model.getConstrs().length);
        
        HashSet<Label> ilpSol = getSelectedLabelsFromIlpSolution(labels, x);
        double wIlp = getObjective(ilpSol, interferences, alpha);
        double weightIlp = getObjectiveWeight(ilpSol);
        writer.println("Objective (test): " + wIlp);
        writer.println("Weight (test): " + weightIlp);
        
        exportLabels("polygons_" + useClique + "_" + instanceID + ".shp", labels, x);
        exportPoints("points_" + useClique + "_" + instanceID + ".shp", points, x);
        exportInterferences("interferences_" + useClique + "_" + instanceID + ".shp", interferencesArray, y);
        
        model.dispose();
        modelLP.dispose();
        
        numInstances++;
    }
    

    private static HashSet<Label> roundingHeuristic(Label[] labels, LinkedList<LinkedList<RectangleWithInfo<Label>>> cliques, GRBVar[] xCont) throws GRBException {

        System.out.println("start rounding " + labels.length + " " + xCont.length);
        
        HashMap<Label, LinkedList<LinkedList<RectangleWithInfo<Label>>>> cliquesByLabel = new HashMap<Label, LinkedList<LinkedList<RectangleWithInfo<Label>>>>();
        for (LinkedList<RectangleWithInfo<Label>> myClique : cliques) {
            for (RectangleWithInfo<Label> rwi : myClique) {
                Label l = rwi.getInfo();
                LinkedList<LinkedList<RectangleWithInfo<Label>>> cliquesForL = cliquesByLabel.get(l);
                if (cliquesForL == null) {
                    cliquesForL = new LinkedList<LinkedList<RectangleWithInfo<Label>>>();
                }
                cliquesForL.add(myClique);
                cliquesByLabel.put(l, cliquesForL );
            }
        }
                
        //sort labels by value in optimal solution to lp relaxation
        TreeMap<Double, LinkedList<Label>> labelsByVal = new TreeMap<Double, LinkedList<Label>>();
        for (int i = 0; i < labels.length; i++) {
            Double d = xCont[i].get(GRB.DoubleAttr.X);
            LinkedList<Label> list = labelsByVal.get(d);
            if (list == null) {
                list = new LinkedList<Label>();
                labelsByVal.put(d, list);               
            }
            list.push(labels[i]);
        }
        
        //add labels if possible
        Quadtree treeLabels = new Quadtree();
        
        HashSet<Label> selection = new HashSet<Label>();
        for (Map.Entry<Double, LinkedList<Label>> me : labelsByVal.descendingMap().entrySet()) {
            for (Label l : me.getValue()) {
                boolean isAllowed = true;
                
                Envelope env = l.getPolygon().getEnvelopeInternal();
                List queryResults = treeLabels.query(env);
                for (Object o : queryResults) {
                    Label l2 = (Label) o;
                    if (l2.getPolygon().overlaps(l.getPolygon())) {
                        isAllowed = false;
                        break;
                    }
                }
                
                if (isAllowed) { //overlap constraint not violated
                                        
                    for (LinkedList<RectangleWithInfo<Label>> myClique : cliquesByLabel.get(l)) {
                        int counter = 0;
                        for (RectangleWithInfo<Label> r : myClique) {
                            if (selection.contains(r.getInfo()) || r.getInfo() == l) {
                                counter++;
                                if (counter > k_gamma) {
                                    isAllowed = false;
                                    break;
                                }
                            }
                        }
                    }
                    
                    if (isAllowed) { //viewbox constraint not violated 
                        treeLabels.insert(l.getPolygon().getEnvelopeInternal(), l);                 
                        selection.add(l);
                    }
                }
            }
        }
        
        System.out.println("end rounding");
        return selection;   
    }

	private static double getObjectiveWeight(HashSet<Label> ilpSol) {
	    double w = 0.0;
        for (Label l : ilpSol) {
            w += l.getWeight();
        }
        return w;
    }

    private static double getObjective(HashSet<Label> roundingSol, TreeMap<Interference, Double> interferences, double alpha) {
		double w = 0.0;
		for (Label l : roundingSol) {
			w += l.getWeight();
		}
		if (interferences != null) {
    		for (Interference c : interferences.keySet()) {
    			if (roundingSol.contains(c.getLabel1()) && roundingSol.contains(c.getLabel2())) {
    				w -= alpha * interferences.get(c);
    			}
    		}
		}
		return w;
	}

	private static HashSet<Label> getSelectedLabelsFromIlpSolution(Label[] labels, GRBVar[] xCont) throws GRBException {
		HashSet<Label> selection = new HashSet<Label>();
		for (int i = 0; i < labels.length; i++) {
			int d = (int) Math.rint( xCont[i].get(GRB.DoubleAttr.X) );
			if (d == 1) selection.add(labels[i]);
		}
		return selection;
	}
	
	private static double[] getRandomWeights(int n, int k, double min, double max, int seed) {
		Random random = new Random(seed);
		double[] w = new double[n * k];
		for (int i = 0; i < n; i++) {
			w[i * k + 3] = w[i * k + 2] = w[i * k + 1] = w[i * k] = min + (max - min) * random.nextDouble(); 
		}
		return w;
	}


	private static Point[] getRandomPointSet(double width, double height, int n, int seed) {
		Random random = new Random(seed);
		Point[] points = new Point[n];
		for (int i = 0; i < n; i++) {
			points[i] = new Point(width * random.nextDouble(), height * random.nextDouble());
		}
		return points;
	}
	
	private static void exportPolygons(String filename, Polygon[] polygons, GRBVar[] x) {
		ShapefileWriter shp_output = new ShapefileWriter();
		File f = new File(filename);
        DriverProperties dpw = new DriverProperties(f.getAbsolutePath());
        try {
            FeatureSchema fs = new FeatureSchema();
            fs.addAttribute("SHAPE", AttributeType.GEOMETRY);
            fs.addAttribute("type", AttributeType.INTEGER);
            LinkedList<BasicFeature> myList = new LinkedList<BasicFeature>();
            
            for (int i = 0; i < polygons.length; i++) {
                Polygon p = polygons[i];
            	BasicFeature bf = new BasicFeature(fs);
                bf.setGeometry(p);
                
                if (x != null) {
                    double d = x[i].get(GRB.DoubleAttr.X);
                    System.out.println("x[" + i + "] = " + d);
                    bf.setAttribute("type", (int) Math.rint(d));
               }
                
                myList.push(bf);
            }
            
            
            FeatureCollection myFeatureCollection = new FeatureDataset(myList, fs);
            System.out.println("Shape written to " + filename);
            shp_output.write(myFeatureCollection, dpw);
            
        } catch(Exception ex) { 
            System.out.println("shp_write: " + ex);
        }
	}
	
	private static void exportPoints(String filename, HashSet<Point> pArr, GRBVar[] x) {
		ShapefileWriter shp_output = new ShapefileWriter();
		File f = new File(filename);
        DriverProperties dpw = new DriverProperties(f.getAbsolutePath());
        
        try {
        	GeometryFactory geomf = new GeometryFactory();
        	
            FeatureSchema fs = new FeatureSchema();
            fs.addAttribute("SHAPE", AttributeType.GEOMETRY);
            fs.addAttribute("type", AttributeType.INTEGER);
            LinkedList<BasicFeature> myList = new LinkedList<BasicFeature>();
            
            for ( Point p : pArr ) {
                BasicFeature bf = new BasicFeature(fs);
				Coordinate[] coords = new Coordinate[1];
				coords[0] = new Coordinate(p.getX(), p.getY());                
            	bf.setGeometry(new com.vividsolutions.jts.geom.Point(new CoordinateArraySequence(coords), geomf));
            	
            	int type = 0;
            	for (Label l: p.getLabels()) {
            		if ((int) Math.rint(x[l.getIndex()].get(GRB.DoubleAttr.X)) == 1) {
            			type = 1;
            			break;
            		}
            	}
            	bf.setAttribute("type", type);
            	
                myList.push(bf);
            }
                        
            FeatureCollection myFeatureCollection = new FeatureDataset(myList, fs);
            System.out.println("Shape written to " + filename);
            shp_output.write(myFeatureCollection, dpw);
            
        } catch(Exception ex) { 
            System.out.println("shp_write: " + ex);
        }
	}
	
	private static void exportPoints(String filename, HashSet<Point> pArr, HashSet<Label> x) {
		ShapefileWriter shp_output = new ShapefileWriter();
		File f = new File(filename);
        DriverProperties dpw = new DriverProperties(f.getAbsolutePath());
        
        try {
        	GeometryFactory geomf = new GeometryFactory();
        	
            FeatureSchema fs = new FeatureSchema();
            fs.addAttribute("SHAPE", AttributeType.GEOMETRY);
            fs.addAttribute("type", AttributeType.INTEGER);
            LinkedList<BasicFeature> myList = new LinkedList<BasicFeature>();
            
            for ( Point p : pArr ) {
                BasicFeature bf = new BasicFeature(fs);
				Coordinate[] coords = new Coordinate[1];
				coords[0] = new Coordinate(p.getX(), p.getY());                
            	bf.setGeometry(new com.vividsolutions.jts.geom.Point(new CoordinateArraySequence(coords), geomf));
            	
            	int type = 0;
            	for (Label l: p.getLabels()) {
            		if (x.contains(l)) {
            			type = 1;
            			break;
            		}
            	}
            	bf.setAttribute("type", type);
            	
                myList.push(bf);
            }
                        
            FeatureCollection myFeatureCollection = new FeatureDataset(myList, fs);
            System.out.println("Shape written to " + filename);
            shp_output.write(myFeatureCollection, dpw);
            
        } catch(Exception ex) { 
            System.out.println("shp_write: " + ex);
        }
	} 
	
	private static void exportInterferences(String filename, Interference[] confArr, GRBVar[] y) {
		if (confArr == null) return;
		
		File f = new File(filename);
		ShapefileWriter shp_output = new ShapefileWriter();
        DriverProperties dpw = new DriverProperties(f.getAbsolutePath());
        
        try {
        	GeometryFactory geomf = new GeometryFactory();
        	
            FeatureSchema fs = new FeatureSchema();
            fs.addAttribute("SHAPE", AttributeType.GEOMETRY);
            fs.addAttribute("type", AttributeType.INTEGER);
            fs.addAttribute("index", AttributeType.INTEGER);
            LinkedList<BasicFeature> myList = new LinkedList<BasicFeature>();
            
            for (int i = 0; i < confArr.length; i++) {
                Interference c = confArr[i];
            	BasicFeature bf = new BasicFeature(fs);
                
            	Point p1 = c.getLabel1().getCenter();
            	Point p2 = c.getLabel2().getCenter();
            	            	
				Coordinate[] coords = new Coordinate[2];
				coords[0] = new Coordinate(p1.getX(), p1.getY());
				coords[1] = new Coordinate(p2.getX(), p2.getY());
				LineString ls = new LineString(new CoordinateArraySequence(coords ), geomf);
                
				bf.setAttribute("type", (int) Math.rint(y[i].get(GRB.DoubleAttr.X) - 0.1));
				bf.setAttribute("index", i);
            	bf.setGeometry(ls);
                myList.push(bf);
            }
                        
            FeatureCollection myFeatureCollection = new FeatureDataset(myList, fs);
            System.out.println("Shape written to " + filename);
            shp_output.write(myFeatureCollection, dpw);
                        
        } catch(Exception ex) { 
            System.out.println("shp_write: " + ex);
        }
	}
	
	
	private static void exportLabels(String filename, Label[] labels, GRBVar[] x) {
	    File f = new File(filename);
		ShapefileWriter shp_output = new ShapefileWriter();
        DriverProperties dpw = new DriverProperties(f.getAbsolutePath());
        try {
            FeatureSchema fs = new FeatureSchema();
            fs.addAttribute("SHAPE", AttributeType.GEOMETRY);
            fs.addAttribute("type", AttributeType.INTEGER);
            fs.addAttribute("weight", AttributeType.DOUBLE);
            fs.addAttribute("name", AttributeType.STRING);
            fs.addAttribute("index", AttributeType.INTEGER);
            LinkedList<BasicFeature> myList = new LinkedList<BasicFeature>();
            
            for (int i = 0; i < labels.length; i++) {
                Polygon p = labels[i].getPolygon();
            	BasicFeature bf = new BasicFeature(fs);
                bf.setGeometry(p);
                double d = x[i].get(GRB.DoubleAttr.X);
                bf.setAttribute("type", (int) Math.rint(d));
                bf.setAttribute("weight", (double) labels[i].getWeight());
                bf.setAttribute("index", i);
                if (labels[i].getText() != null) bf.setAttribute("name", labels[i].getText());
                myList.push(bf);
            }
                        
            FeatureCollection myFeatureCollection = new FeatureDataset(myList, fs);
            System.out.println("Shape written to " + filename);
            shp_output.write(myFeatureCollection, dpw);
            
        } catch(Exception ex) { 
            System.out.println("shp_write: " + ex);
        }
	}
		
	private static void exportLabels(String filename, Label[] labels, HashSet<Label> x) {
		File f = new File(filename);
	    ShapefileWriter shp_output = new ShapefileWriter();
        DriverProperties dpw = new DriverProperties(f.getAbsolutePath());
        try {
            FeatureSchema fs = new FeatureSchema();
            fs.addAttribute("SHAPE", AttributeType.GEOMETRY);
            fs.addAttribute("type", AttributeType.INTEGER);
            fs.addAttribute("weight", AttributeType.DOUBLE);
            fs.addAttribute("name", AttributeType.STRING);
            fs.addAttribute("index", AttributeType.INTEGER);
            LinkedList<BasicFeature> myList = new LinkedList<BasicFeature>();
            
            for (int i = 0; i < labels.length; i++) {
                Polygon p = labels[i].getPolygon();
            	BasicFeature bf = new BasicFeature(fs);
                bf.setGeometry(p);
                if (x != null && x.contains(labels[i])) {
                	bf.setAttribute("type", 1);	
                } else {
                	bf.setAttribute("type", 0);	
                }
                bf.setAttribute("index", i);
                bf.setAttribute("weight", (double) labels[i].getWeight());
                if (labels[i].getText() != null) bf.setAttribute("name", labels[i].getText());
                myList.push(bf);
            }
                        
            FeatureCollection myFeatureCollection = new FeatureDataset(myList, fs);
            System.out.println("Shape written to " + filename);
            shp_output.write(myFeatureCollection, dpw);
            
        } catch(Exception ex) { 
            System.out.println("shp_write: " + ex);
        }
	}
	
}
	
