package rectangleSweep;

import java.util.ArrayList;
import java.util.Collections;
import java.util.LinkedList;
import java.util.SortedSet;
import java.util.TreeSet;

/**
 * @author haunert
 * 
 * When using or adapting this code for your research, please cite:
 * 
 * Jan-Henrik Haunert and Alexander Wolff
 * Beyond Maximum Independent Set: An Extended Integer Programming Formulation for Point Labeling
 * International Journal of Geo-Information, 2017
 */
public class RectangleSweep<E> {
    private LinkedList<RectangleWithInfo<E>> rectangles;
    private static boolean verbose = false;
    public RectangleSweep(LinkedList<RectangleWithInfo<E>> ll) {
        rectangles = ll;
    }
    
    public LinkedList<LinkedList<RectangleWithInfo<E>>> computeMaximalCliques() {
        //Event queue for sweep from left to right 
        ArrayList<Event<E>> queue = new  ArrayList<Event<E>>(2 * rectangles.size());
        for (RectangleWithInfo<E> r : rectangles) {
           queue.add(new Event<E>(r, true, r.getxMin()));
           queue.add(new Event<E>(r, false, r.getxMin() + r.getWidth()));
        }
        Collections.sort(queue);
                
        //vertical status line containing intersection points with intersected rectangles
        TreeSet<Event<E>> status = new TreeSet<Event<E>>();
        //all cliques that have been found so far, each represented as a list of rectangles
        LinkedList<LinkedList<RectangleWithInfo<E>>> cliques = new LinkedList<LinkedList<RectangleWithInfo<E>>>();
        
        //process events from left to right
        for (Event<E> e : queue) {
           if (verbose) System.out.println("e:" + e.isBegin() + " " + e.getR().getName());
           if (e.isBegin()) { //a new rectangle begins
              //create two event points on status line (lower and upper boundary of rectangle)
              Event<E> e1 = new Event<E>(e.getR(), true, e.getR().getyMin());
              e.getR().setBeginOnStatus(e1);
              status.add(e1);
              Event<E> e2 = new Event<E>(e.getR(), false, e.getR().getyMin() + e.getR().getHeight());
              e.getR().setEndOnStatus(e2);
              status.add(e2);             
           } else { // a rectangle r ends
               if (verbose) {
                   String sstring = "status: ";
                   for (Event<E> etemp : status) {
                       sstring += etemp.toShortString() + ";";
                   }
                   System.out.println(sstring);
               } 
               
               
              //get all event points on status line between lower and upper boundary of rectangle
              Event<E> e1 = e.getR().getBeginOnStatus();
              Event<E> e2 = e.getR().getEndOnStatus();
              SortedSet<Event<E>> subset = status.subSet(e1, e2);
              
              //the queue containing a set of rectangles that mutually intersect
              LinkedList<RectangleWithInfo<E>> rectQueue = new LinkedList<RectangleWithInfo<E>>();
              
              //add all rectangles to queue that are below r and intersect r on status line
              for (Event<E> e3 : subset) {
                 if (!e3.isBegin()) {
                    rectQueue.addLast(e3.getR());
                 }
              }
              //process queue
              for (Event<E> e3 :subset) {
                 if (e3.isBegin()) { //add rectangle that begins
                    rectQueue.addLast(e3.getR());
                 } else { //rectangle ends -> generate new clique                     
                    LinkedList<RectangleWithInfo<E>> clique = new LinkedList<RectangleWithInfo<E>>();
                    clique.addAll(rectQueue);
                    cliques.addLast(clique);
                    
                    if (verbose) {
                        String cstring = "clique: ";
                        for (RectangleWithInfo<E> rwi : clique) cstring += rwi.getName() + " ";
                        System.out.println(cstring);
                    }
                    
                    rectQueue.removeFirst();
                 }
              }
              LinkedList<RectangleWithInfo<E>> lastClique = new LinkedList<RectangleWithInfo<E>>();
              lastClique.addAll(rectQueue);
              cliques.addLast(lastClique);
              
              if (verbose) {
                  String cstring = "clique: ";
                  for (RectangleWithInfo<E> rwi : lastClique) cstring += rwi.getName() + " ";
                  System.out.println(cstring);
              }
              
              status.remove(e1);
              status.remove(e2);
           }
        }
        return cliques;
     }
}
